/*
 * rcelf - Enable RAMCrypt for a given ELF file
 *
 * Copyright (C) 2015 Johannes Goetzfried <johannes.goetzfried@cs.fau.de>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307 USA.
 */

#include <elf.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define RAMCRYPT_ELF_FLAG 0x10


static void usage(char *name)
{
	printf("Usage: %s [-s|-c] <binary>\n", name);
	printf("Shows the RAMCrypt status for the given binary.\n");
	printf("    -s: Enable RAMCrypt for the given binary.\n");
	printf("    -c: Disable RAMCrypt for the given binary.\n");
	exit(EXIT_FAILURE);
}


int main(int argc, char **argv)
{
	int set_bit   = 0;
	int clear_bit = 0;
	int elf64;
	long offset;
	char *file;
	char *mode;
	FILE* fp;

	unsigned char ehdr[sizeof (Elf64_Ehdr)];
	unsigned char phdr[sizeof (Elf64_Phdr)];
	Elf32_Ehdr *ehdr32 = (Elf32_Ehdr *)ehdr;
	Elf32_Phdr *phdr32 = (Elf32_Phdr *)phdr;
	Elf64_Ehdr *ehdr64 = (Elf64_Ehdr *)ehdr;
	Elf64_Phdr *phdr64 = (Elf64_Phdr *)phdr;
	Elf32_Word p_flags;

	unsigned long phentsize;
	unsigned long phnum;
	unsigned long i;

	if (argc < 2 || argc > 3)
		usage(*argv);

	file = argv[1];
	if (argc == 3) {
		file = argv[2];
		if (strcmp(argv[1], "-s") == 0)
			set_bit = 1;
		else if (strcmp(argv[1], "-c") == 0)
			clear_bit = 1;
		else
			usage(*argv);
	}

	mode = (set_bit || clear_bit) ? "r+" : "r";
	if ((fp = fopen(file, mode)) == NULL) {
		perror("Could not open file");
		return EXIT_FAILURE;
	}

	if (fread(ehdr, sizeof (Elf64_Ehdr), 1, fp) < 1) {
		fprintf(stderr, "Not an ELF file!\n");
		return EXIT_FAILURE;
	}

	if (memcmp(ehdr32->e_ident, ELFMAG, SELFMAG) != 0) {
		fprintf(stderr, "Not an ELF file!\n");
		return EXIT_FAILURE;
	}

	if (ehdr32->e_ident[EI_CLASS] == ELFCLASS32)
		elf64 = 0;
	else if (ehdr32->e_ident[EI_CLASS] == ELFCLASS64)
		elf64 = 1;
	else {
		fprintf(stderr, "No valid ELF file!\n");
		return EXIT_FAILURE;
	}

	offset = elf64 ? ehdr64->e_phoff : ehdr32->e_phoff;
	if (fseek(fp, offset, SEEK_SET) < 0) {
		perror("Could not seek to Phdr");
		return EXIT_FAILURE;
	}

	phentsize = elf64 ? ehdr64->e_phentsize : ehdr32->e_phentsize;
	if (phentsize > sizeof (Elf64_Phdr)) {
		fprintf(stderr, "Invalid Phentsize!\n");
		return EXIT_FAILURE;
	}

	phnum = elf64 ? ehdr64->e_phnum : ehdr32->e_phnum;
	for (i = 0; i < phnum; i++) {
		if (fread(phdr, phentsize, 1, fp) < 1) {
			fprintf(stderr, "Could not read Phdr!\n");
			return EXIT_FAILURE;
		}
		if (phdr32->p_type == PT_GNU_STACK)
			break;
	}

	if (i == phnum) {
		fprintf(stderr, "ELF has no PT_GNU_STACK Phdr!\n");
		return EXIT_FAILURE;
	}

	p_flags = elf64 ? phdr64->p_flags : phdr32->p_flags;

	if (!set_bit && !clear_bit) {
		printf("Flags: %08x\n", p_flags);
		printf("RAMCrypt is%s activated!\n",
			(p_flags & RAMCRYPT_ELF_FLAG) ? "" : " not");
		fclose(fp);
		return EXIT_SUCCESS;
	}

	if (set_bit)
		p_flags |= RAMCRYPT_ELF_FLAG;
	else
		p_flags &= ~RAMCRYPT_ELF_FLAG;

	if (elf64)
		phdr64->p_flags = p_flags;
	else
		phdr32->p_flags = p_flags;

	if (fseek(fp, offset + i*phentsize, SEEK_SET) < 0) {
		perror("Could not seek to Phdr");
		return EXIT_FAILURE;
	}

	if (fwrite(phdr, phentsize, 1, fp) < 1) {
		fprintf(stderr, "Could not write Phdr!\n");
		return EXIT_FAILURE;
	}

	fclose(fp);
	return EXIT_SUCCESS;
}
